@@ -13,10 +13,7 @@ in MySQL specific way. The module uses DBD/DBI interface; whenever
 possible it is recommended to use direct MySQL module that works
 directly with database without DBD/DBI layer in between.
 
-This is the lowest level XAO::FS knows about. Underneath of it are
-DBD::mysql, DBI, database itself, operationg system, hardware, atoms,
-protons, gluons and so on and on and on.. The level might be not so low
-if we look at it this way.
+This is the lowest level XAO::FS knows about.
 
 =head1 METHODS
 
@@ -27,15 +24,14 @@ if we look at it this way.
 ###############################################################################
 package XAO::DO::FS::Glue::MySQL_DBI;
 use strict;
+use Error qw(:try);
 use XAO::Utils qw(:debug :args :keys);
 use XAO::Objects;
-use DBI;
-use DBD::mysql;
 
-use base XAO::Objects->load(objname => 'Atom');
+use base XAO::Objects->load(objname => 'FS::Glue::SQL_DBI');
 
 use vars qw($VERSION);
-($VERSION)=(q$Id: MySQL_DBI.pm,v 1.9 2002/03/19 03:37:08 am Exp $ =~ /(\d+\.\d+)/);
+($VERSION)=(q$Id: MySQL_DBI.pm,v 1.15 2002/10/29 09:23:59 am Exp $ =~ /(\d+\.\d+)/);
 
 ###############################################################################
 
@@ -59,30 +55,21 @@ sub new ($%) {
 
     my $self=$proto->SUPER::new($args);
 
-    ##
-    # Connecting to the database
-    #
-    my $user=$args->{user};
-    my $password=$args->{password};
     my $dsn=$args->{dsn};
     $dsn || $self->throw("new - required parameter missed 'dsn'");
     $dsn=~/^OS:(\w+):(\w+)(;.*)?$/ || $self->throw("new - bad format of 'dsn' ($dsn)");
-    $1 eq 'MySQL_DBI' || $self->throw("new - driver type is not MySQL_DBI");
+    my $driver=$1;
     my $dbname=$2;
     my $dbopts=$3 || '';
-    my $dbh=DBI->connect("DBI:mysql:$dbname$dbopts",$user,$password) ||
-            $self->throw("new - can't connect to the database ($dsn,$user,***)");
-    $self->{dbh}=$dbh;
-    my $v=$dbh->{mysql_serverinfo} || $dbh->{serverinfo};
-    if(!$v || $v !~ /^(\d+)\.(\d+)(\.(\d+))?$/ || $1<3 || $2<23) {
-        $self->{no_null_indexes}=1;
-        eprint "Disabling NULL indexes, older MySQL found";
-    }
 
-    ##
-    # Returning resulting object
-    #
-    $self;
+    $driver =~ '^MySQL' ||
+        throw $self "new - wrong driver type ($driver)";
+
+    $self->sql_connect(dsn => "DBI:mysql:$dbname$dbopts",
+                       user => $args->{user},
+                       password => $args->{password});
+
+    return $self;
 }
 
 ###############################################################################
@@ -91,7 +78,7 @@ sub new ($%) {
 
 Adds new integer field to the given table. First parameter is table
 name, then field name, then index flag, then unique flag, then minimal
-value and then maximum value.
+value and then maximum value and default value.
 
 B<Note:> Indexes only work with MySQL 3.23 and later.
 
@@ -99,7 +86,7 @@ B<Note:> Indexes only work with MySQL 3.23 and later.
 
 sub add_field_integer ($$$$$) {
     my $self=shift;
-    my ($table,$name,$index,$unique,$min,$max,$connected)=@_;
+    my ($table,$name,$index,$unique,$min,$max,$default,$connected)=@_;
     $name.='_';
 
     $min=-0x80000000 unless defined $min;
@@ -138,25 +125,23 @@ sub add_field_integer ($$$$$) {
         }
     }
 
-    $sql.=' NOT NULL' if $self->{no_null_indexes};
+    $sql.=" NOT NULL DEFAULT $default";
 
     $sql="ALTER TABLE $table ADD $name $sql";
 
-    $self->{dbh}->do($sql) || $self->throw_sql('add_field_integer');
+    $self->sql_do($sql);
 
     if(($index || $unique) && (!$unique || !$connected)) {
         my $usql=$unique ? " UNIQUE" : "";
         $sql="ALTER TABLE $table ADD$usql INDEX fsi__$name ($name)";
         #dprint ">>>$sql<<<";
-        $self->{dbh}->do($sql) ||
-            $self->throw_sql('add_field_text');
+        $self->sql_do($sql);
     }
 
     if($unique && $connected) {
         $sql="ALTER TABLE $table ADD UNIQUE INDEX fsu__$name (parent_unique_id_,$name)";
         #dprint ">>>$sql<<<";
-        $self->{dbh}->do($sql) ||
-            $self->throw_sql('add_field_text');
+        $self->sql_do($sql);
     }
 }
 
@@ -166,7 +151,7 @@ sub add_field_integer ($$$$$) {
 
 Adds new real field to the given table. First parameter is table name,
 then field name, then index flag, then unique flag, then optional
-minimal value and then optional maximum value.
+minimal value and then optional maximum value and default value.
 
 B<Note:> Indexes only work with MySQL 3.23 and later.
 
@@ -174,28 +159,24 @@ B<Note:> Indexes only work with MySQL 3.23 and later.
 
 sub add_field_real ($$$;$$) {
     my $self=shift;
-    my ($table,$name,$index,$unique,$min,$max,$connected)=@_;
+    my ($table,$name,$index,$unique,$min,$max,$default,$connected)=@_;
     $name.='_';
 
-    my $sql="ALTER TABLE $table ADD $name DOUBLE";
+    my $sql="ALTER TABLE $table ADD $name DOUBLE NOT NULL DEFAULT $default";
 
-    $sql.=' NOT NULL' if $self->{no_null_indexes};
-
-    $self->{dbh}->do($sql) || $self->throw_sql('add_field_real');
+    $self->sql_do($sql);
 
     if(($index || $unique) && (!$unique || !$connected)) {
         my $usql=$unique ? " UNIQUE" : "";
         $sql="ALTER TABLE $table ADD$usql INDEX fsi__$name ($name)";
         #dprint ">>>$sql<<<";
-        $self->{dbh}->do($sql) ||
-            $self->throw_sql('add_field_text');
+        $self->sql_do($sql);
     }
 
     if($unique && $connected) {
         $sql="ALTER TABLE $table ADD UNIQUE INDEX fsu__$name (parent_unique_id_,$name)";
         #dprint ">>>$sql<<<";
-        $self->{dbh}->do($sql) ||
-            $self->throw_sql('add_field_text');
+        $self->sql_do($sql);
     }
 }
 
@@ -204,9 +185,9 @@ sub add_field_real ($$$;$$) {
 =item add_field_text ($$$$$)
 
 Adds new text field to the given table. First is table name, then field
-name, then index flag, then unique flag, maximum length and 'connected'
-flag. Depending on maximum length it will create CHAR, TEXT, MEDIUMTEXT
-or LONGTEXT.
+name, then index flag, then unique flag, maximum length, default value
+and 'connected' flag. Depending on maximum length it will create CHAR,
+TEXT, MEDIUMTEXT or LONGTEXT.
 
 'Connected' flag must be set if that table holds elements deeper into
 the tree then the top level.
@@ -218,7 +199,7 @@ later.
 
 sub add_field_text ($$$$$$) {
     my $self=shift;
-    my ($table,$name,$index,$unique,$max,$connected)=@_;
+    my ($table,$name,$index,$unique,$max,$default,$connected)=@_;
     $name.='_';
 
     my $sql;
@@ -232,12 +213,8 @@ sub add_field_text ($$$$$$) {
         $sql="LONGTEXT";
     }
 
-    $sql.=' NOT NULL' if $self->{no_null_indexes};
-
-    #dprint ">>ALTER TABLE $table ADD $name $sql<<";
-
-    $self->{dbh}->do("ALTER TABLE $table ADD $name $sql") ||
-        throw_sql $self 'add_field_text';
+    $self->sql_do("ALTER TABLE $table ADD $name $sql NOT NULL DEFAULT ?",
+                  $default);
 
     !$unique || $max<=255 ||
         throw $self "add_field_text - property is too long to make it unique ($max)";
@@ -248,15 +225,13 @@ sub add_field_text ($$$$$$) {
         my $usql=$unique ? " UNIQUE" : "";
         $sql="ALTER TABLE $table ADD$usql INDEX fsi__$name ($name)";
         #dprint ">>>$sql<<<";
-        $self->{dbh}->do($sql) ||
-            $self->throw_sql('add_field_text');
+        $self->sql_do($sql);
     }
 
     if($unique && $connected) {
         $sql="ALTER TABLE $table ADD UNIQUE INDEX fsu__$name (parent_unique_id_,$name)";
         #dprint ">>>$sql<<<";
-        $self->{dbh}->do($sql) ||
-            $self->throw_sql('add_field_text');
+        $self->sql_do($sql);
     }
 }
 
@@ -284,7 +259,7 @@ sub add_table ($$$$) {
                                  : "") .
             ")";
 
-    $self->{dbh}->do($sql) || $self->throw_sql('add_table');
+    $self->sql_do($sql);
 }
 
 ###############################################################################
@@ -299,9 +274,7 @@ sub delete_row ($$$) {
     my $self=shift;
     my ($table,$uid)=@_;
 
-    my $sth=$self->{dbh}->prepare("DELETE FROM $table WHERE unique_id=?");
-    $sth && $sth->execute(''.$uid) && $sth->finish ||
-        $self->throw_sql('drop_row');
+    $self->sql_do("DELETE FROM $table WHERE unique_id=?",$uid);
 }
 
 ###############################################################################
@@ -314,13 +287,7 @@ will do that for you.
 =cut
 
 sub disconnect ($) {
-    my $self=shift;
-    my $dbh=$self->{dbh};
-	if($dbh) {
-        $self->unlock_tables();
-        $dbh->disconnect();
-        delete $self->{dbh};
-    }
+    shift->sql_disconnect;
 }
 
 ###############################################################################
@@ -344,19 +311,16 @@ sub drop_field ($$$$$$) {
     if($index && (!$unique || !$connected)) {
         my $sql="ALTER TABLE $table DROP INDEX fsi__$name";
         # dprint ">>>$sql<<<";
-        $self->{dbh}->do($sql) ||
-            throw_sql $self 'drop_field';
+        $self->sql_do($sql);
     }
 
     if($unique && $connected) {
         my $sql="ALTER TABLE $table DROP INDEX fsu__$name";
         # dprint ">>>$sql<<<";
-        $self->{dbh}->do($sql) ||
-            throw_sql $self 'drop_field';
+        $self->sql_do($sql);
     }
 
-    $self->{dbh}->do("ALTER TABLE $table DROP $name") ||
-        throw_sql $self 'drop_field';
+    $self->sql_do("ALTER TABLE $table DROP $name");
 }
 
 ###############################################################################
@@ -372,8 +336,32 @@ sub drop_table ($$) {
     my $self=shift;
     my $table=shift;
 
-    my $sth=$self->{dbh}->prepare("DROP TABLE $table");
-    $sth && $sth->execute && $sth->finish || $self->throw_sql('drop_table');
+    $self->sql_do("DROP TABLE $table");
+}
+
+###############################################################################
+
+=item increment_key_seq ($)
+
+Increments the value of key_seq in Global_Fields table identified by the
+given row unique ID. Returns previous value.
+
+=cut
+
+sub increment_key_seq ($$) {
+    my $self=shift;
+    my $uid=shift;
+
+    my $sth=$self->sql_execute('SELECT key_seq_ FROM Global_Fields WHERE unique_id=?',$uid);
+    my $seq=$self->sql_first_row($sth)->[0];
+    if(!$seq) {
+        $self->sql_do('UPDATE Global_Fields SET key_seq_=2 WHERE unique_id=?',$uid);
+        return 1;
+    }
+    else {
+        $self->sql_do('UPDATE Global_Fields SET key_seq_=key_seq_+1 WHERE unique_id=?',$uid);
+        return $seq;
+    }
 }
 
 ###############################################################################
@@ -387,53 +375,40 @@ objects database.
 
 sub initialize_database ($) {
     my $self=shift;
-    my $dbh=$self->{dbh};
 
-    my $sth=$dbh->prepare('SHOW TABLES');
-    $sth && $sth->execute() || throw_sql $self 'initialize_database';
-    my %tables;
-    while(my ($table)=$sth->fetchrow_array()) {
-        $tables{$table}=1;
-        $dbh->do("DROP TABLE $table") || throw_sql $self 'initialize_database';
+    my $sth=$self->sql_execute('SHOW TABLES');
+    while(my $row=$self->sql_fetch_row($sth)) {
+        $self->sql_do("DROP TABLE $row->[0]");
     }
-    $sth->finish;
+    $self->sql_finish($sth);
 
     my @initseq=(
         <<'END_OF_SQL',
 CREATE TABLE Global_Fields (
-  unique_id INT unsigned NOT NULL AUTO_INCREMENT,
-  table_name_ CHAR(30) NOT NULL default '',
-  field_name_ CHAR(30) NOT NULL default '',
-  type_ CHAR(20) NOT NULL default '',
-  refers_ CHAR(30) default NULL,
-  index_ TINYINT default NULL,
-  default_ CHAR(30) default NULL,
-  maxlength_ INT unsigned default NULL,
-  maxvalue_ DOUBLE default NULL,
-  minvalue_ DOUBLE default NULL,
+  unique_id INT UNSIGNED NOT NULL AUTO_INCREMENT,
+  table_name_ CHAR(30) NOT NULL DEFAULT '',
+  field_name_ CHAR(30) NOT NULL DEFAULT '',
+  type_ CHAR(20) NOT NULL DEFAULT '',
+  refers_ CHAR(30) DEFAULT NULL,
+  key_format_ CHAR(100) DEFAULT NULL,
+  key_seq_ INT UNSIGNED DEFAULT NULL,
+  index_ TINYINT DEFAULT NULL,
+  default_ CHAR(30) DEFAULT NULL,
+  maxlength_ INT UNSIGNED DEFAULT NULL,
+  maxvalue_ DOUBLE DEFAULT NULL,
+  minvalue_ DOUBLE DEFAULT NULL,
   PRIMARY KEY  (table_name_,field_name_),
   UNIQUE KEY unique_id (unique_id)
 )
 END_OF_SQL
         <<'END_OF_SQL',
 INSERT INTO Global_Fields VALUES (1,'Global_Data','project',
-                                  'text','',0,'',40,NULL,NULL)
-END_OF_SQL
-        <<'END_OF_SQL',
-CREATE TABLE Global_Dictionary (
-  table_name varchar(20) NOT NULL default '',
-  field_name varchar(20) NOT NULL default '',
-  strip varchar(100) NOT NULL default '',
-  table_uid int(10) unsigned NOT NULL default '0',
-  INDEX (table_name,field_name),
-  INDEX (strip(8)),
-  INDEX (table_uid)
-)
+                                  'text','',NULL,NULL,0,'',40,NULL,NULL)
 END_OF_SQL
         <<'END_OF_SQL',
 CREATE TABLE Global_Data (
   unique_id INT UNSIGNED NOT NULL AUTO_INCREMENT,
-  project_ char(40),
+  project_ char(40) NOT NULL DEFAULT '',
   PRIMARY KEY (unique_id)
 )
 END_OF_SQL
@@ -443,19 +418,19 @@ END_OF_SQL
         <<'END_OF_SQL',
 CREATE TABLE Global_Classes (
   unique_id int(10) unsigned NOT NULL AUTO_INCREMENT,
-  class_name_ char(100) NOT NULL default '',
-  table_name_ char(30) NOT NULL default '',
+  class_name_ char(100) NOT NULL DEFAULT '',
+  table_name_ char(30) NOT NULL DEFAULT '',
   PRIMARY KEY  (unique_id),
   UNIQUE KEY  (class_name_)
 )
 END_OF_SQL
         <<'END_OF_SQL',
-INSERT INTO Global_Classes VALUES (2,'FS::Global','Global_Data')
+INSERT INTO Global_Classes VALUES (1,'FS::Global','Global_Data')
 END_OF_SQL
     );
 
     foreach my $clause (@initseq) {
-        $dbh->do($clause) || throw_sql $self 'initialize_database';
+        $self->sql_do($clause);
     }
 }
 
@@ -477,16 +452,14 @@ sub list_keys ($$$$$) {
 
     my $sth;
     if($conn_name) {
-        $sth=$self->{dbh}->prepare("SELECT $key FROM $table" .
-                                   " WHERE $conn_name=?");
-        $sth->execute($conn_value) || $self->throw_sql('list_keys');
+        $sth=$self->sql_execute("SELECT $key FROM $table WHERE $conn_name=?",
+                                $conn_value);
     }
     else {
-        $sth=$self->{dbh}->prepare("SELECT $key FROM $table");
-        $sth->execute() || $self->throw_sql();
+        $sth=$self->sql_execute("SELECT $key FROM $table");
     }
 
-    [ map { $_->[0] } @{$sth->fetchall_arrayref([0])} ];
+    return $self->sql_first_column($sth);
 }
 
 ###############################################################################
@@ -507,21 +480,32 @@ and store them would do the job?
 sub load_structure ($) {
     my $self=shift;
 
+    ##
+    # Checking if Global_Fields table has key_format_ and key_seq_
+    # fields. Adding them if it does not.
+    #
+    my $sth=$self->sql_execute("DESC Global_Fields");
+    my $flist=$self->sql_first_column($sth);
+    if(! grep { $_ eq 'key_format_' } @$flist) {
+        $self->sql_do('ALTER TABLE Global_Fields ADD key_format_ CHAR(100) DEFAULT NULL');
+    }
+    if(! grep { $_ eq 'key_seq_' } @$flist) {
+        $self->sql_do('ALTER TABLE Global_Fields ADD key_seq_ INT UNSIGNED DEFAULT NULL');
+    }
+
     ##
     # Loading fields descriptions from the database.
     #
     my %fields;
-    my $dbh=$self->{dbh};
-    my $sth=$dbh->prepare("SELECT table_name_,field_name_,type_," .
-                                 "refers_,index_,default_," .
-                                 "maxlength_,minvalue_,maxvalue_" .
-                          " FROM Global_Fields");
-    $sth && $sth->execute() ||
-        throw_sql $self 'load_structure';
+    $sth=$self->sql_execute("SELECT unique_id,table_name_,field_name_," .
+                                   "type_,refers_,key_format_," .
+                                   "index_,default_," .
+                                   "maxlength_,minvalue_,maxvalue_" .
+                            " FROM Global_Fields");
 
-    while(my ($table,$field,$type,
-              $refers,$index,$default,
-              $maxlength,$minvalue,$maxvalue)=$sth->fetchrow_array) {
+    while(my $row=$self->sql_fetch_row($sth)) {
+        my ($uid,$table,$field,$type,$refers,$key_format,
+            $index,$default,$maxlength,$minvalue,$maxvalue)=@$row;
         my $data;
         if($type eq 'list') {
             $refers || $self->throw("load_structure - no class name at Global_Fields($table,$field,..)");
@@ -534,7 +518,9 @@ sub load_structure ($) {
             $refers || $self->throw("load_structure - no class name at Global_Fields($table,$field,..)");
             $data={
                 type        => $type,
-                refers      => $refers
+                refers      => $refers,
+                key_format  => $key_format,
+                key_unique_id => $uid,
             };
         }
         elsif($type eq 'connector') {
@@ -568,23 +554,24 @@ sub load_structure ($) {
         }
         $fields{$table}->{$field}=$data;
     }
-    $sth->finish();
+    $self->sql_finish($sth);
 
     ##
     # Now loading classes translation table and putting fields
     # descriptions inside of it as well.
     #
-    $sth=$dbh->prepare("SELECT class_name_,table_name_ FROM Global_Classes");
-    $sth && $sth->execute() || $self->throw("load_structure - SQL error");
+    $sth=$self->sql_execute("SELECT class_name_,table_name_ FROM Global_Classes");
     my %classes;
-    while(my ($class,$table)=$sth->fetchrow_array) {
+    while(my $row=$self->sql_fetch_row($sth)) {
+        my ($class,$table)=@$row;
         my $f=$fields{$table};
         $f || $self->throw("load_structure - no description for $table table (class $class)");
-        $classes{$class}={ table => $table,
-                           fields => $f
-                         };
+        $classes{$class}={
+            table   => $table,
+            fields  => $f,
+        };
     }
-    $sth->finish();
+    $self->sql_finish($sth);
 
     ##
     # Resulting structure
@@ -618,7 +605,7 @@ it.
 
 =cut
 
-sub retrieve_fields ($$$$) {
+sub retrieve_fields ($$$@) {
     my $self=shift;
     my $table=shift;
     my $unique_id=shift;
@@ -630,13 +617,8 @@ sub retrieve_fields ($$$$) {
     my $sql=join(',',@names);
     $sql="SELECT $sql FROM $table WHERE unique_id=?";
 
-    my $sth=$self->{dbh}->prepare($sql);
-    $sth && $sth->execute($unique_id) || $self->throw_sql("retrieve_field");
-
-    my $row=$sth->fetchrow_arrayref;
-    $sth->finish;
-
-    $row;
+    my $sth=$self->sql_execute($sql,$unique_id);
+    return $self->sql_first_row($sth);
 }
 
 ###############################################################################
@@ -653,16 +635,28 @@ sub search ($%) {
     my $self=shift;
     my $query=get_args(\@_);
 
-    my $sth=$self->{dbh}->prepare($query->{sql});
-    $sth && $sth->execute(@{$query->{values}}) || $self->throw_sql('search');
+    my $sql=$query->{sql};
 
-    my @results;
-    while(my @row=$sth->fetchrow_array) {
-        push @results,\@row;
+    if($query->{options} && $query->{options}->{limit}) {
+        $sql.=' LIMIT '.int($query->{options}->{limit});
+    }
+
+    # dprint "SQL: $sql";
+
+    my $sth=$self->sql_execute($sql,$query->{values});
+
+    if(scalar(@{$query->{fields_list}})>1) {
+        my @results;
+        while(my $row=$self->sql_fetch_row($sth)) {
+            push @results,$row;
+        }
+        $self->sql_finish($sth);
+        return \@results;
+    }
+    else {
+        return $self->sql_first_column($sth);
     }
-    $sth->finish;
 
-    \@results;
 }
 
 ###############################################################################
@@ -701,19 +695,6 @@ sub search_clause_ws ($$$) {
 
 ###############################################################################
 
-=item setup_dictionary ($$$)
-
-Supposed to set up dictionary tables if required. For MySQL driver it
-does nothing as all dictionaries are stored in the same table currently
-and this table is created when initial database layout is created.
-
-=cut
-
-sub setup_dictionary ($$$$) {
-}
-
-###############################################################################
-
 =item store_row ($$$$$$$)
 
 Stores complete row of data into the given table. New name is generated
@@ -737,28 +718,39 @@ sub store_row ($$$$$$$) {
     $key_name.='_';
     $conn_name.='_' if $conn_name;
 
-    $self->lock_tables($table);
+    ##
+    # We need to lock Global_Fields too as it might be used in AUTOINC
+    # key formats.
+    #
+    my @ltab=$table eq 'Global_Fields' ? ('Global_Fields') : ($table,'Global_Fields');
+    $self->lock_tables(@ltab);
 
     my $uid;
-    if($key_value) {
+    if(ref($key_value) eq 'CODE') {
+        my $kv;
+        while(1) {
+            $kv=&{$key_value};
+            last unless $self->unique_id($table,
+                                         $key_name,$kv,
+                                         $conn_name,$conn_value,
+                                         1);
+        }
+        $key_value=$kv;
+    }
+    elsif($key_value) {
         $uid=$self->unique_id($table,
                               $key_name,$key_value,
                               $conn_name,$conn_value,
                               1);
-    } else {
-        while(1) {
-            $key_value=generate_key();
-            #$key_value=sprintf('%5.2f',rand(100));
-            last unless $self->unique_id($table,
-                                         $key_name,$key_value,
-                                         $conn_name,$conn_value,
-                                         1);
-        }
     }
-
+    else {
+        throw $self "store_row - no key_value given (old usage??)";
+    }
+    
     if($uid) {
         $self->update_row($table,$uid,$row);
-    } else {
+    }
+    else {
         my @fn=($key_name, map { $_.'_' } keys %{$row});
         my @fv=($key_value, values %{$row});
         if($conn_name && $conn_value) {
@@ -772,11 +764,10 @@ sub store_row ($$$$$$$) {
         $sql.=join(',',('?') x scalar(@fn));
         $sql.=')';
 
-        my $sth=$self->{dbh}->prepare($sql);
-        $sth && $sth->execute(@fv) || $self->throw_sql('store_row');
+        $self->sql_do($sql,\@fv);
     }
 
-    $self->unlock_tables($table);
+    $self->unlock_tables(@ltab);
 
     $key_value;
 }
@@ -798,75 +789,16 @@ sub unique_id ($$$$$$$) {
 
     my $sth;
     if(defined($conn_name) && defined($conn_value)) {
-        $sth=$self->{dbh}->prepare("SELECT unique_id FROM $table WHERE $conn_name=? AND $key_name=?");
-        $sth && $sth->execute(''.$conn_value,''.$key_value) ||
-            $self->throw_sql("unique_id");
-    } else {
-        $sth=$self->{dbh}->prepare("SELECT unique_id FROM $table WHERE $key_name=?");
-        $sth && $sth->execute(''.$key_value) ||
-            $self->throw_sql("unique_id");
+        $sth=$self->sql_execute("SELECT unique_id FROM $table WHERE $conn_name=? AND $key_name=?",
+                                ''.$conn_value,''.$key_value);
+    }
+    else {
+        $sth=$self->sql_execute("SELECT unique_id FROM $table WHERE $key_name=?",
+                                ''.$key_value);
     }
 
-    my $row=$sth->fetchrow_arrayref;
-    $sth->finish;
-    $row ? $row->[0] : undef;
-}
-
-###############################################################################
-
-=item update_dictionary ($table $uid $name $value)
-
-Updates dictionary for the given field. Dictionary is supported by two
-tables Global_Dictionary and Global_Backrefs. The first table is
-dictionary itself, while the second holds references that allow to
-delete/modify records in the dictionary quicker.
-
-Here is an example content of Global_Dictionary with two names ('John
-Silver' and 'John Doe') encoded. It assumes that unique_id's of rows
-that holds `John Silver' and 'John Doe' in Customers table are 15 and 25
-respectfully.
-
- | unique_id | table_name | field_name | strip  |  ids  |
- +-----------+------------+------------+--------+-------+
- |         1 | Customers  | name       | john   | 15,25 |
- |         2 | Customers  | name       | silver | 15    |
- |         3 | Customers  | name       | doe    | 25    |
-
-Structure of Global_Backrefs for the same data:
-
- | unique_id | table_name | table_uid | field_name | ids |
- +-----------+------------+-----------+------------+-----+
- |         1 | Customers  | 15        | name       | 1,2 |
- |         2 | Customers  | 25        | name       | 1,3 |
-
-=cut
-
-sub update_dictionary ($$$$$) {
-    my $self=shift;
-    my ($table,$uid,$name,$strips)=@_;
-    $name.='_';
-
-    $self->lock_tables('Global_Dictionary');
-
-    my %rows;
-    my $dbh=$self->{dbh};
-    my $sth=$dbh->prepare('DELETE FROM Global_Dictionary' .
-                          ' WHERE table_name=? AND field_name=? AND table_uid=?');
-    $sth && $sth->execute($table,$name,$uid) ||
-        $self->throw_sql('update_dictionary');
-
-    $sth=$dbh->prepare('INSERT INTO Global_Dictionary' .
-                       ' (table_name,field_name,strip,table_uid)' .
-                       ' VALUES (?,?,?,?)');
-    my %inserted;
-    foreach my $strip (@{$strips}) {
-        next if $inserted{$strip};
-        $inserted{$strip}=1;
-        $sth && $sth->execute($table,$name,$strip,$uid) ||
-            $self->throw_sql('update_dictionary');
-    }
-
-    $self->unlock_tables();
+    my $row=$self->sql_first_row($sth);
+    return $row ? $row->[0] : undef;
 }
 
 ###############################################################################
@@ -882,51 +814,17 @@ Stores new value into single data field. Example:
 sub update_field ($$$$$) {
     my $self=shift;
     my ($table,$unique_id,$name,$value)=@_;
+
+    $unique_id ||
+        throw $self "update_field($table,..,$name,..) - no unique_id given";
+
     $name.='_';
-    $unique_id || $self->throw("update_field($table,..,$name,..) - no unique_id given");
-    my $sth=$self->{dbh}->prepare("UPDATE $table SET $name=? WHERE unique_id=?");
-    $sth && $sth->execute(defined($value) ? ''.$value : undef,''.$unique_id) ||
-        $self->throw_sql("update_field");
-}
 
-###############################################################################
+    defined($value) ||
+        throw $self "update_field($table,..,$name,..) - undefined value given";
 
-=item update_key ($$$$) {
-
-Stores new value into key field in the given table. If value for key is
-not given then it generates new random one just like store_row does.
-
- $self->_driver->update_key($table,$unique_id,$key_name,$key_value);
-
-=cut
-
-sub update_key ($$$$$) {
-    my $self=shift;
-    my ($table,$unique_id,$key_name,$key_value,$conn_name,$conn_value)=@_;
-    $key_name.='_';
-    $conn_name.='_' if $conn_name;
-
-    $self->lock_tables($table);
-
-    if(! $key_value) {
-        while(1) {
-            $key_value=generate_key();
-            last unless $self->unique_id($table,
-                                         $key_name,$key_value,
-                                         $conn_name,$conn_value,
-                                         1);
-        }
-    }
-
-    $self->update_field($table,$unique_id,$key_name,$key_value);
-
-    if(defined($conn_name) && defined($conn_value)) {
-        $self->update_field($table,$unique_id,$conn_name,$conn_value);
-    }
-
-    $self->unlock_tables();
-
-    return $key_value;
+    $self->sql_do("UPDATE $table SET $name=? WHERE unique_id=?",
+                  ''.$value,$unique_id);
 }
 
 ###############################################################################
@@ -945,13 +843,13 @@ sub update_row ($$$$) {
     my $self=shift;
     my ($table,$uid,$row)=@_;
 
+    return unless keys %$row;
+
     my $sql="UPDATE $table SET ";
     $sql.=join(',',map { "${_}_=?" } keys %{$row});
     $sql.=' WHERE unique_id=?';
 
-    my $sth=$self->{dbh}->prepare($sql);
-    $sth && $sth->execute(values %{$row},$uid) ||
-        $self->throw_sql('update_row');
+    my $sth=$self->sql_do($sql,values %$row,$uid);
 }
 
 ###################################################################### PRIVATE
@@ -960,14 +858,13 @@ sub lock_tables ($@) {
     my $self=shift;
     my $sql='LOCK TABLES ';
     $sql.=join(',',map { "$_ WRITE" } @_);
-    $self->{dbh}->do($sql) || $self->throw_sql('lock_tables');
+    $self->sql_do($sql);
 }
 
 sub unlock_tables ($) {
     my $self=shift;
-    return unless $self->{dbh};
-    $self->{dbh}->do('UNLOCK TABLES') ||
-        die 'unlock_tables - failed';
+    return unless $self->sql_connected;
+    $self->sql_do_no_error('UNLOCK TABLES');
 }
 
 sub throw ($@) {
@@ -976,15 +873,9 @@ sub throw ($@) {
     $self->SUPER::throw(@_);
 }
 
-sub throw_sql ($$) {
-    my $self=shift;
-    my $method=shift;
-    $self->throw($method . ' - SQL error (' . $self->{dbh}->errstr . ')');
-}
-
 sub DESTROY ($) {
     my $self=shift;
-	$self->disconnect();
+	$self->sql_disconnect();
 }
 
 ###############################################################################
@@ -995,13 +886,16 @@ __END__
 
 =head1 AUTHORS
 
-Xao, Inc. (c) 2001. This module was developed by Andrew Maltsev
-<am@xao.com> with help and valuable comments from other team members.
+Copyright (c) 2001,2002 XAO, Inc.
+
+This module was developed by Andrew Maltsev <am@xao.com> with the help
+and valuable comments from other team members.
 
 =head1 SEE ALSO
 
 Further reading:
 L<XAO::FS>,
+L<XAO::DO::FS::Glue::SQL_DBI>,
 L<XAO::DO::FS::Glue>.
 
 =cut
