@@ -67,7 +67,7 @@ use XAO::Objects;
 use base XAO::Objects->load(objname => 'FS::Glue');
 
 use vars qw($VERSION);
-($VERSION)=(q$Id: Hash.pm,v 1.5 2002/03/19 03:37:08 am Exp $ =~ /(\d+\.\d+)/);
+($VERSION)=(q$Id: Hash.pm,v 1.9 2002/10/29 17:50:29 am Exp $ =~ /(\d+\.\d+)/);
 
 ###############################################################################
 
@@ -132,6 +132,56 @@ When possible it is recommended to name the key as the class name or
 some derivative from it with '_id' suffix. If you add Orders list
 placeholder it is recommended to call the key 'order_id'.
 
+=item key_format
+
+Only for 'list' placeholders. Specifies format of the auto-generated
+keys for a single argument put() method on Lists (L<XAO::DO::FS::List>).
+
+Key format consists of normal alpha-numeric characters and may include
+several special sequences:
+
+=over
+
+=item <$RANDOM$>
+
+Random 8-character sequence as generated by XAO::Utils::generate_key()
+function. It never evaluates to 'false' in perl sense, always starts
+from a letter and has some other constraints - see L<XAO::Utils> for
+details.
+
+
+=item <$AUTOINC$>
+
+This is an auto-incrementing unsigned at least 32-bit integer value. It
+is not guaranteed to be continuously incrementing, but new value is
+always greater then any previously assigned and greater then zero. There
+is no definition of what would happen if all integer values are used up.
+
+Can optionally have number of digits specified - <$AUTOINC/10$> to
+generate 10 digits integers. Default is to use minimum possible number
+of digits.
+
+=item <$GMTIME$>
+
+Seconds GMT since Epoch.
+
+=item <$DATE$>
+
+Current date and time in YYYYMMDDHHMMSS format.
+
+=back
+
+One of <$AUTOINC$> or <$RANDOM$> must present in the key format as
+otherwise it is impossible to guarantee key uniqueness.
+
+The default key format is '<$RANDOM$>'.
+
+Examples:
+
+ V<$AUTOINC$>        - V1, V2, V3, ..., V12345
+ <$DATE$>_<$RANDOM$> - 200210282218_QWERT123
+ NUM<$AUTOINC/12$>   - NUM000000000001, ..., NUM000000012345
+
 =item maxlength
 
 Maximum length for 'text' type in characters. If later on you will try
@@ -281,6 +331,7 @@ sub build_structure ($%) {
             while(my ($n,$v)=each %ph) {
                 next if $n eq 'name';
                 next if $n eq 'key';
+                next if $n eq 'key_format';
                 next if $n eq 'structure';
                 my $dbv=$desc->{$n};
                 (!defined($v) && !defined($dbv)) || $dbv eq $v ||
@@ -913,11 +964,13 @@ sub put ($$$) {
         }
     }
     elsif($type eq 'integer' || $type eq 'real') {
+        $value=$self->_field_default($name,$field) if $value eq '';
+
         !defined($field->{minvalue}) || $value>=$field->{minvalue} ||
-            $self->throw("put - value ($value) is bigger then $field->{minvalue} for $name");
+            $self->throw("put - value ($value) is less then $field->{minvalue} for $name");
 
         !defined($field->{maxvalue}) || $value<=$field->{maxvalue} ||
-            $self->throw("put - value ($value) is less then $field->{maxvalue} for $name");
+            $self->throw("put - value ($value) is bigger then $field->{maxvalue} for $name");
 
         if($$self->{detached}) {
             $$self->{data}->{$name}=$value;
